#include "lcd.h"

// timpii (ms) cat ii ia LCD-ctrl sa faca init, fetch instr si fetch data
#define init_time 20
#define instr_time 10
#define data_time 10

int LCD_is_on = 1;

void LCD_clear(){
	// instructiune clear
	PORTD = 1 << PB1;
}

void LCD_init(){
	// setare functie: 11 caractere / linie, 2 linii, 8 biti interfata, cursor
	PORTD = 0x3c;
	
	_delay_ms(init_time);
	
	// setare mod: cursor inaintare in sens +, display shifting
	PORTD = 0x07;
	
	// scrie 0
	LCD_write_char('0');
}

unsigned int char_to_lcd_data(char c){
	return (unsigned int)c;
}

void LCD_write_char(char c){
	// afisare caracter pe LCD
	// write data to DDRAM
	_delay_ms(data_time);
	
	PORTB |= (1 << RS);
	PORTB &= ~(1 << RW);
	PORTD = char_to_lcd_data(c);
}

void LCD_write_string(char* s){
	// afisare string pe LCD
	// setare DDRAM
	_delay_ms(instr_time);
	
	PORTB &= ~(1 << RS);
	PORTB &= ~(1 << RW);
	PORTD = 0xff;
	
	for (; s; s++){
		LCD_write_char(*s);
		
		// astept o ms dupa un caracter trimis
		_delay_ms(1);
	}
}
		

void LCD_output(double current_value, double l10p){
	// afisare rezultat pe LCD
	// l10p: cea mai mica putere a lui 10 (corespunde ultimei cifre)
	LCD_clear();
	
	int current_int = (int)current_value;
	while (current_int > 0){
		LCD_write_char(key_to_char(current_int % 10));
		current_int /= 10;
	}
	int current_frac = (current_value - (int)current_value) / l10p;
	if (current_frac > 0){
		LCD_write_char('.');
		while (current_frac > 0){
			LCD_write_char(key_to_char(current_frac % 10));
			current_int /= 10;
		}
	}
}

void LCD_error(char *c){
	LCD_clear();
	LCD_write_string(c);	
}
 
void LCD_on(){
	// setare DB3 (PD4 la PORTD)
	PORTD |= 1 << DB3;
	LCD_is_on = 1;
	
	_delay_ms(instr_time);
}

void LCD_off(){
	// resetare PB4 -> resetare V0 la LCD
	PORTD &= ~(1 << DB3);
	LCD_is_on = 0;
	
	_delay_ms(instr_time);
}

void LCD_switch(){
	if (LCD_is_on) LCD_on();
	else LCD_off();
}