#include <avr\io.h>
#include <avr\interrupt.h>
#include <util/delay.h>

#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "lcd.h"
#include "utils.h"

#define F_CPU 16000000 // frecventa proc
#define INTR_TIME 500 // timp intrerupere
#define PI 3.14
#define EPS 1e-8
#define precision 20 // precizia (count) pentru calcul prin dezvoltare in serie

// valoarea curenta (pentru output)
double current_value;

// expresia de evaluat
char expr[256];
int length = 0; // lungime expresie

// cea mai nesemnificativa cifra, trunchiata la putere a lui 10
double last_10_power = 1.0; 


// functii redefinite
double _pow(double x, double y){
	return exp(x*log(y));	
}

double _sqrt(double x){
	return pow(x, 0.5);
}


int check(char expr[256], int start, int end){
	// verifica corectitudinea expresiei expr (de la start la end exclusiv)
	int i;
	
	// sa nu inceapa cu semne, in afara de -
	if (expr[start] < 0x30 && expr[i] != '-')
		return 0;
		
	// sa nu fie 2 semne consectutive, in afara de "op(" sau ")op"
	for (i = start; i < end; i++){
		if (expr[i] < 0x30 && expr[i+1]  < 0x30 && 
			(!(expr[i] > 0x29 /*op*/ && expr[i+1] == '(')) && 
			(!(expr[i+1] > 0x29 /*op*/ && expr[i] == ')')))
				return 0;
	}
	
	return 1;
	
}


int eval(char expr[256], int start, int end, double *result){
	// evalueaza expresia (din expr) (de la start la end exclusiv)	
	// rezultatul este scris in result
	// intoarce 0 in caz de succes sau != 0 in caz de exceptie 
	// (/0, ln(<0), sqrt(<0), etc)
	
	
	// cautare +
	// gasire -> separare in subexpresii
	int i;
	double *rez;
	rez = 0;
	
	for (i = start; i < end; i++)
		if (expr[i] == '+'){
			int e1 = eval(expr, start, i, result);
			if (e1 > 0)	return e1;
			
			int e2 = eval(expr, start, i, rez);
			if (e2 > 0)	return e2;
			
			*result += *rez;
		}
			
	// cautare -
	// gasire -> separare in subexpresii
	for (i = start; i < end; i++)
		if (expr[i] == '-'){
			int e1 = eval(expr, start, i, result);
			if (e1 > 0)	return e1;
			
			int e2 = eval(expr, start, i, rez);
			if (e2 > 0)	return e2;
			
			*result -= *rez;
		}
			
	// cautare *
	// gasire -> separare in subexpresii
	for (i = start; i < end; i++)
		if (expr[i] == '*'){
			int e1 = eval(expr, start, i, result);
			if (e1 > 0)	return e1;
			
			int e2 = eval(expr, start, i, rez);
			if (e2 > 0)	return e2;
			
			*result *= *rez;
		}
			
	// cautare /
	// gasire -> separare in subexpresii
	for (i = start; i < end; i++)
		if (expr[i] == '/'){
			int e1 = eval(expr, start, i, result);
			if (e1 > 0)	return e1;
			
			int e2 = eval(expr, start, i, rez);
			if (e2 > 0)	return e2;
			if (rez == 0) return 1; // impartire la 0
			
			*result /= *rez;
		}
			
	// cautare ^
	// gasire -> separare in subexpresii
	for (i = start; i < end; i++)
		if (expr[i] == '^'){
			int e1 = eval(expr, start, i, result);
			if (e1 > 0)	return e1;
			
			int e2 = eval(expr, start, i, rez);
			if (e2 > 0)	return e2;
			
			if (*result == 0 && *rez <= 0) 
				return 2; // 0^(<=0)
			if (*result < 1 && (int)(*rez) != (*rez)) // radacina (nu neaparat sqrt) din nr < 0
				return 2; 
			
			*result = _pow(*result, *rez);
		}
			
	
	// paranteze la inceput sau sfarsit -> eliminare
	if (expr[start] == '(')
		return eval(expr, start+1, end, result);
	if (expr[end] == ')')
		return eval(expr, start, end-1, result);
			
	
	// cautare functie
	// gasire -> aplicare
	// gasire -> separare in subexpresii
	double (*f)(double) = NULL;
	for (i = start; i < end; i++){
		if (!strncmp(expr+i, "R", 1)){
			//f = _sqrt;
			break;
		}
		if (!strncmp(expr+i, "exp", 3)){
			//f = exp;
			break;
		}
		if (!strncmp(expr+i, "ln", 2)){
			//f = log;
			break;
		}
		if (!strncmp(expr+i, "sin", 3)){
			//f = sin;
			break;
		}
		if (!strncmp(expr+i, "cos", 3)){
			//f = cos;
			break;
		}
		if (!strncmp(expr+i, "tan", 3)){
			//f = tan;
			break;
		}
		if (!strncmp(expr+i, "asin", 4)){
			//f = asin;
			break;
		}
		if (!strncmp(expr+i, "acos", 4)){
			//f = acos;
			break;
		}
		if (!strncmp(expr+i, "atan", 4)){
			//f = atan;
			break;
		}
	}
			
	
	// calcul efectiv
	*result = 0;
	
	// cifrele dinainte de virgula
	for (i = start; i < end && expr[i] != '.'; i++)
		*result = *result * 10 + (expr[i] - 0x30) % 10;
	
	// cifrele de dupa virgula
	// cea mai mica putere a lui 10, corespunzatoare ultimei cifre
	double l10p = 0.1; 
	for (; i < end; i++)
		*result += ((expr[i] - 0x30) % 10) * (l10p /= 10);
	
	if (f) {
		if (f == _sqrt && *result < 0)
			return 3; // radical din nr < 0
		if (f == log && *result < 0) 
			return 3; // eroare ln
		if (f == tan && (int)(*result) % 180 == 90)
			return 3; // eroare tangenta
		if ((f == asin || f == acos) && (*result < -1 || *result > 1))
			return 3; // eroare arcsin / arccos
			
		if (f == sin || f == cos || f == tan)
			*result *= PI/180.0;
		
		*result = f(*result);
		
		if (f == asin || f == acos || f == atan)
			*result *= 180.0/PI;
	}
	
	return 0;
}

void clear(){
	// resetare valoare, expresie
	current_value = 0;
	memset(expr, 0, 256);
	LCD_clear();
	LCD_write_char('0');
}

char key_to_char(int key){
	// converteste un cod cheie la caracterul corespunzator
	if (key < 11)
		return (key < 10 ? key : 0) + 0x30;
	switch (key){
		case OP_PLUS: return '+'; 
		case OP_MINUS: return '-'; 
		case OP_MUL: return '*'; 
		case OP_DIV: return '/'; 
		case OP_POW: return '^'; 
		
		case LEFT_P: return '('; 
		case RIGHT_P: return ')'; 
		
		case DECIMAL_SEP: return '.';
	}
}

char* key_to_string(int key){
	// converteste un cod cheie la textul corespunzator
	switch(key){
		case F_SQRT: return "R";
		case F_EXP: return "exp";
		case F_LN: return "ln";
		case F_SIN: return "sin";
		case F_COS: return "cos";
		case F_ASIN: return "asin";
		case F_ACOS: return "acos";
		case F_TAN: return "tan";
		case F_ATAN: return "atan";
		case F_SGN: return "-";
	}
}


void process_key(int key){
	// in functie de tasta apasata key,
	// output sau calcul expresie
	
	if (key == EQUAL){
		// evaluare expresie
		if (check(expr, 0, length)){
			int e = eval(expr, 0, length, &current_value);
			if (e == 0) LCD_output(current_value, EPS);
			else if (e == 1) LCD_error("Impartire la 0");
			else if (e == 2) LCD_error("Eroare ridicare la putere");
			else if (e == 3) LCD_error("Eroare argument functie");
		}
		else LCD_error("Eroare sintaxa");
	}
	else if (key == CLEAR){
		// clear
		clear();
	}
	else if (key == OFF_ON){
		// turn off / on LCD
		LCD_switch();
	}
	else {
		// incrementare lungime expresie
		length++;
		
		// output
		if (key < 18) // cifra sau operator -> un singur caracter
			LCD_write_char(key_to_char(key)); 
		else  // functie -> mai multe caractere
			LCD_write_string(key_to_string(key)); 
	}
}


int get_key_id(){
	// citire tasta apasata
	int i, j;
	
	// prima tastatura
	for (i = 0; i < 4; i++) 
		if (PORTA & (1 << i)){
			// pinul i de pe portA este setat 
			// ->este apasata o tasta de pe linia i
			for (j = 0; j < 4; j++) 
				if (PORTA & (1 << (j+4)))
					// pinul j+4 de pe portA este setat
					// -> este apasata o tasta de pe colana j
					return i*4 + j + 1;
			return NONE;
		}
		
	// a 2-a tastatura
	for (i = 0; i < 4; i++) 
		if (PORTC & (1 << i)){
			// pinul i de pe portA este setat 
			// ->este apasata o tasta de pe linia i
			for (j = 0; j < 4; j++) 
				if (PORTC & (1 << (j+4)))
					// pinul j+4 de pe portA este setat
					// -> este apasata o tasta de pe colana j
					return i*4 + j + 17;
			return NONE;
		}
					
}

int main(){
	int key;
 
	LCD_on();
 
	// initializare tastatura 1 (aritmetica)
	DDRA = 0x00;		// setare PA ca input
	PORTA = 0x00; 		
 
	// initializare tastatura 2 (scientific)
	DDRC = 0x00;		// setare PC ca input
	PORTC = 0x00;
  
	// init LCD (instructiuni)
	DDRB |= 0x0F;		// setare PB(4:1) ca output LCD(3:6)
	PORTB = 0x00;
	PORTB |= 1 << PB4;
 
	// init LCD (date)
	DDRD |= 0x8F;		// setare PD(7:1) ca output LCD(8:14)
	PORTD = 0x00;	
	
	LCD_init(); // date de init LCD
 
	// bucla principala
	while (1){
		_delay_ms(INTR_TIME);
		key = get_key_id(); // primim id tasta apasata (1-32)
		if (key > 0){
			process_key(key);	         // procesam caracterul
		}
	}
  
	return 0;
}